def spec_line(data, label, color):
    return {'data': data, 'label': label, 'color': color}


def spec_axis(title, lines):
    return {'title': title, 'lines': lines}


def spec_chart(title, axes):
    return {'title': title, 'axes': axes}


def experimental_with_ratio_spec(metro, complexity):
    return spec_chart(
        '~ Elapsed time {} in seconds'.format(metro.elapsed_time),
        [
            spec_axis(
                'Experimental',
                [
                    spec_line(metro.ratio, r'ratio: $\mathcal{T}/\mathcal{O}$', 'tab:green'),
                    spec_line(metro.experimental, r'experimental: $\mathcal{T}(\mathcal{f}(n))$', 'tab:blue'),
                ]
            ),
            spec_axis(
                'Theoretical',
                [
                    spec_line(metro.theoretical, complexity.latex, 'tab:red')
                ]
            )
        ]
    )



import math


def loop_n(instance=[]):
    n = len(instance)
    j = 0
    for _ in range(n):
        j += 1
    return j


def loop_n_quadratic(instance=[]):
    n = len(instance)
    j = 0
    for _ in range(n):
        for _ in range(n):
            j += 1
    return j


def loop_n_cubic(instance=[]):
    n = len(instance)
    j = 0
    for _ in range(n):
        for _ in range(n):
            for _ in range(n):
                j += 1
    return j


def loop_n_log(instance=[]):
    n = len(instance)
    log_n = math.ceil(math.log2(n))
    j = 0
    for _ in range(log_n):
        j += 1
    return j


def loop_n_log_n(instance=[]):
    n = len(instance)
    log_n = math.ceil(math.log2(n))
    j = 0
    for _ in range(n * log_n):
        j += 1
    return j



import almetro.chart as chart
import almetro.table as table
from almetro.complexity import cn
from copy import deepcopy
import timeit


class Metro:
    def __init__(self, complexity):
        self.__start = timeit.default_timer()
        self.__elapsed_time = 0
        self.__experimental = {}
        self.__theoretical = {}
        self.__ratio = {}
        self.__complexity = complexity
        self.__processed = False

    @property
    def elapsed_time(self):
        return self.__elapsed_time

    @property
    def experimental(self):
        return deepcopy(self.__experimental)

    @property
    def theoretical(self):
        return deepcopy(self.__theoretical)

    @property
    def ratio(self):
        return deepcopy(self.__ratio)

    @property
    def complexity(self):
        return self.__complexity

    @property
    def processed(self):
        return self.__processed

    def register(self, size, timestats):
        self.__elapsed_time = timeit.default_timer() - self.__start
        self.__experimental[size] = min(timestats)
        self.__processed = False

    def chart(self):
        self.process()
        return chart.build_chart(self)

    def table(self):
        self.process()
        return table.build_table(self)

    def process(self):
        if not self.__processed:
            if not self.__complexity:
                raise ValueError('complexity must be provided')
            self.__theoretical = {}
            self.__ratio = {}
            for n, t_fn in self.__experimental.items():
                self.__theoretical[n] = self.__complexity.fn(n)
                self.__ratio[n] = t_fn / max(self.__theoretical[n], 0.1)
            self.__processed = True

    @staticmethod
    def new(complexity=cn):
        return Metro(complexity)



from almetro.instance import growing
from almetro.metro import Metro
import timeit


class ExecutionSettings:
    def __init__(self, trials=1, runs=1):
        if not trials or trials < 1:
            raise TypeError('#trials must be provided')
        if not runs or runs < 1:
            raise TypeError('#runs must be provided')
        self.trials = trials
        self.runs = runs

    @staticmethod
    def new():
        return ExecutionSettings()


class InstanceSettings:
    def __init__(self, instances=1, provider=growing()):
        if not instances:
            raise TypeError('#instances must be provided')
        if not provider:
            raise TypeError('provider must be provided')
        self.instances = instances
        self.provider = provider

    @staticmethod
    def new():
        return InstanceSettings()


class Al:

    def __init__(self, instance_settings=InstanceSettings.new(), execution_settings=ExecutionSettings.new()):
        if not instance_settings:
            raise TypeError('instance settings must be provided')
        if not execution_settings:
            raise TypeError('execution settings must be provided')
        self.__instance_settings = instance_settings
        self.__execution_settings = execution_settings

    def with_instances(self, instances, provider):
        return Al(instance_settings=InstanceSettings(instances, provider), execution_settings=self.__execution_settings)

    def with_execution(self, trials, runs):
        return Al(instance_settings=self.__instance_settings, execution_settings=ExecutionSettings(trials, runs))

    def metro(self, algorithm, complexity):
        metro = Metro.new(complexity)
        for _ in range(self.__instance_settings.instances):
            instance = self.__instance_settings.provider.new_instance()

            def runner():
                algorithm(**instance)
            metro.register(instance['size'], timeit.repeat(runner, number=self.__execution_settings.runs, repeat=self.__execution_settings.trials))
        return metro



from matplotlib import pyplot as plt
from collections.abc import Iterable
from .spec import experimental_with_ratio_spec


class ChartLine:
    def __init__(self, data={}, label=None, color=None):
        self.data = data
        self.label = label
        self.color = color
        self.plt_line = None

    def setup(self, plt_axis):
        lines = plt_axis.plot(self.data.keys(), self.data.values(), self.color, label=self.label)
        if isinstance(lines, Iterable):
            self.plt_line = lines[0]
        else:
            self.plt_line = lines

    @staticmethod
    def new(data={}, label=None, color=None):
        return ChartLine(data, label, color)


class ChartAxis:
    legend_loc = "upper left"
    grid = True

    def __init__(self, title=None, chart_lines=[]):
        self.title = title
        self.chart_lines = chart_lines
        self.plt_axis = None

    def setup(self, plt_axis):
        self.plt_axis = plt_axis
        self.plt_axis.set_title(self.title)
        self.plt_axis.grid(ChartAxis.grid)
        for chart_line in self.chart_lines:
            chart_line.setup(self.plt_axis)
        self.plt_axis.legend(loc=ChartAxis.legend_loc)
        self.plt_axis.label_outer()

    @staticmethod
    def new(title=None, chart_lines=[]):
        return ChartAxis(title, chart_lines)


class Chart:
    def __init__(self, title=None, chart_axes=[]):
        self.title = title
        self.chart_axes = chart_axes
        self.plt_figure = None

    def show(self):
        fig, axes = plt.subplots(1, len(self.chart_axes), sharex=True, sharey=False, figsize=(18, 10))
        self.plt_figure = fig
        self.plt_figure.suptitle(self.title)
        for i, chart_axis in enumerate(self.chart_axes):
            if isinstance(axes, Iterable):
                axis = axes[i]
            else:
                axis = axes
            chart_axis.setup(axis)
        plt.close()
        return fig

    @staticmethod
    def new(title=None, chart_axes=[]):
        return Chart(title, chart_axes)


def build_chart(metro, spec_builder=experimental_with_ratio_spec):
    spec = spec_builder(metro, metro.complexity)
    chart_axes = []
    for axis_spec in spec['axes']:
        chart_lines = []
        for line_spec in axis_spec['lines']:
            chart_lines.append(ChartLine.new(**line_spec))
        chart_axes.append(ChartAxis.new(axis_spec['title'], chart_lines))

    return Chart.new(spec['title'], chart_axes)



import random


class InstanceProvider:

    def __init__(self, fn=None):
        """Provides a static instance creator.

        :param fn fn: A function that returns the instance
        """
        self.fn = fn

    def new_instance(self):
        """Provides a new instance value from fn.

        :returns: any type of instance
        """
        return self.fn()


class GrowingNumberSequenceProvider(InstanceProvider):

    def __init__(self, initial_size=10, growth_rate=0.1, growth_size=0):
        """Provides a growing int sequence without repetition instance creator.

        :param int initial_size: number of elements in the instance. Default is 10.
        :param int growth_rate: growth rate of instance per iteration. Default is 0.1 (10%).
        :param int growth_rate: growth initial_size of instance per iteration. Default is 0.
        """
        self.iterations = 0
        self.initial_size = initial_size
        self.growth = growth_size if growth_size > 0 else initial_size * growth_rate

    def new_instance(self):
        """Provides a new int list as an instance.

        :returns: one int list. For example: [1, 3]
        :rtype: list(int)
        """
        size = int(self.initial_size + (self.iterations * self.growth))
        self.iterations += 1
        return {'instance': random.sample(range(size * 2), size), 'size': size}


class GeneratorProvider(InstanceProvider):

    def new_instance(self):
        """Provides a new instance value from next(fn).

        :returns: dict
        """
        return next(self.fn())


def generator(fn=None):
    return GeneratorProvider(fn)


def growing(initial_size=10, growth_rate=0.1, growth_size=0):
    return GrowingNumberSequenceProvider(initial_size=initial_size, growth_rate=growth_rate, growth_size=growth_size)



import tabulate


def build_table(metro, spec_builder=None):
    data = []
    for n, cost in metro.experimental.items():
        data.append((n, cost, metro.theoretical[n], metro.ratio[n]))

    return tabulate.tabulate(
        data,
        headers=['n', 'experimental', f'theoretical {metro.complexity.text}', 'ratio'],
        tablefmt="fancy_grid"
    )



#!/usr/bin/env python
from setuptools import find_packages, setup

project = "almetro"
version = "1.0.2"

setup(
    name=project,
    version=version,
    license="Apache License",
    description="A python library to measure algorithms execution time and compare with its theoretical complexity",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author="Arnour Sabino",
    author_email="arnour.sabino@gmail.com",
    url="https://github.com/arnour/almetro",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "matplotlib==3.0.3",
        "tabulate==0.8.3"
    ],
    setup_requires=[
        "nose>=1.3.7",
    ],
    dependency_links=[
    ],
    entry_points={
    },
    extras_require={
        "test": [
            "flake8>=3.7.7",
            "flake8-print>=3.1.0",
            "coverage>=4.5.3",
            "PyHamcrest>=1.9.0",
        ],
    },
)



import math


class Complexity:
    def __init__(self, fn=lambda n: n, text='O(n)', latex=r'$\mathcal{O}(n)$'):
        """Provides a complexity notation.

        :param fn fn: A function that calculates the cost
        :param string text: Description label
        :param string latex: Description label compatible with latex notation
        """
        self.__fn = fn
        self.__text = text
        self.__latex = latex

    @property
    def text(self):
        return self.__text

    @property
    def latex(self):
        return self.__latex if self.__latex else self.__text

    @property
    def fn(self):
        return self.__fn


cn = Complexity()
cn_quadratic = Complexity(lambda n: n**2, text='O(n^2)', latex=r'$\mathcal{O}(n^2)$')
cn_cubic = Complexity(lambda n: n**3, text='O(n^3)', latex=r'$\mathcal{O}(n^3)$')
cn_log_n = Complexity(lambda n: n * math.log2(n), text='O(n lg n)', latex=r'$\mathcal{O}(n*\log{n})$')
clog_n = Complexity(lambda n: math.log2(n), text='O(lg n)', latex=r'$\mathcal{O}(\log{n})$')
c_one = Complexity(lambda n: 1, text='O(1)', latex=r'$\mathcal{O}(1)$')
